#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <string.h>

#define DEBUG

#define PAES_LARGE 100000
// objective  functions (BOTH TO BE MINIMIZED) names
#define _LENGTH 0
#define _URGENCYLOSS 1

// solution (beacons tour) represented by a chromosome
typedef struct {
	int *chrom;  // chrom[i] = b if beacon b 
		     // is the ith one visited
	double *obj; // obj[i] is the i'th objective 
		     // function value (_LENGTH and
		     //  _URGENCYLOSS) for solution 
	int grid_loc; // PAES internals
} solution_t;

// Beacons Problem Data
typedef struct {
	int nbBeacons;
	int *urgency; // urgency[b] = v  if the urgency associated to beacon b
	int **distances; // distance matrix for beacons
	double *lat, *lon; // latitude and longitude of points (optional, for drawing only)
} beacons_t;

// PAES params (PAES is the optimization algorithm)
// main parameters read from paes.dat, provided on command line
typedef struct {
	int gens;    // num of iterations (generations)
	int depth;   // internal parameter, e.g. 2
	int objs;    // number of objective functions. Set to 2 at init
	solution_t **archive; // array of solutions that are kept as beeing good trade-offs
	int maxArchive;  // max Size of archive 
	int inArchive;   // current size of archive
	double *gl_offset;   // the range, offset etc of the grid. 3 objectives max 
	double *gl_range;    // internals
	double *gl_largest;  // internals
	int *grid_pop;   // the array holding the population residing in each grid location
} paes_params_t;


/* init funcs */
solution_t *paes_init(char *paramfile, paes_params_t **ppaes, beacons_t *beacons);
beacons_t *read_beacons(char *beaconsfile);

solution_t *new_sol(beacons_t *beacons, paes_params_t *paes); // create and initialize randomly a solution
int evaluate_sol(solution_t * sol, beacons_t *beacons); // evaluate a solution
solution_t *mutate_sol(solution_t * sol, beacons_t *beacons, paes_params_t *paes); // mutate a solution
solution_t *dup_sol(solution_t * sol, beacons_t *beacons, paes_params_t *paes); // copy a solution
solution_t *localsearch_sol(solution_t *sol, beacons_t *beacons, paes_params_t *paes); // 2-opt on a solution
void print_solFP(FILE *fp, solution_t * sol, beacons_t *beacons, paes_params_t *paes); // pront a solution
#define print_sol(sol, beacons, paes) print_solFP(stdout, sol, beacons, paes)
void free_sol(solution_t * sol); // free memory associated to a solution
// draw a map with the solution, only the map if sol==NULL, default filename beacons.pdf if fname==NULL
void draw_sol(solution_t *sol, beacons_t *beacons, paes_params_t *paes, char *fname);

// paes internal functions
void paes_update_grid(solution_t *s, paes_params_t *paes);
void paes_archive_soln(solution_t *s, beacons_t *beacons, paes_params_t *paes);
int paes_compare_to_archive(solution_t *s, paes_params_t *paes);
int compare_min(double *first, double *second, int n);

// for execution time measurement
long msectime();

// initialization:
// * read PAES alg parameters from file
// * read Beacons tour problem data from file 
// initial the PAES structure (initial solution and try to fill the archive with random solutions -- optional)
solution_t *paes_init(char *paramfile, paes_params_t **ppaes, beacons_t *beacons)
{
	int i, ok;
	solution_t *current, *first, *second;
	paes_params_t *paes;
     	FILE *fp;

	// for the PAES
	int seed;
	paes = (paes_params_t *)malloc(sizeof(paes_params_t));

	/* reading parameter file with parameters for selection */
	fp = fopen(paramfile, "r"); 
	assert(fp != NULL);

	paes->objs = 2;
	paes->inArchive = 0;
	ok = fscanf(fp, "seed %d\n", &seed);
	assert(ok == 1);
	srand(seed);
	ok = fscanf(fp, "gens %d\n", &(paes->gens));
	assert(ok == 1);
	ok = fscanf(fp, "archive %d\n", &(paes->maxArchive));
	assert(ok == 1);
	ok = fscanf(fp, "depth %d", &(paes->depth));
	assert(ok == 1);
	fclose(fp);


	paes->archive = (solution_t **)malloc(paes->maxArchive*sizeof(solution_t));
	bzero(paes->archive, paes->maxArchive*sizeof(solution_t *));
	paes->gl_offset = (double *)malloc(paes->objs*sizeof(double));   // the range, offset etc of the grid 
	bzero(paes->gl_offset, paes->objs*sizeof(double));
	paes->gl_range = (double *)malloc(paes->objs*sizeof(double));   // the range, offset etc of the grid 
	bzero(paes->gl_range, paes->objs*sizeof(double));
	paes->gl_largest = (double *)malloc(paes->objs*sizeof(double));   // the range, offset etc of the grid 
	bzero(paes->gl_largest, paes->objs*sizeof(double));
	paes->grid_pop = (int *)malloc(1000000*sizeof(int));   // the array holding the population residing in each grid location
	bzero(paes->grid_pop, 1000000*sizeof(int));

	if(!paes->archive) {
                perror("Out of memory. Aborting.");
                exit(-1);
        }

	first = new_sol(beacons, paes);
	paes_update_grid(first, paes);
	paes_archive_soln(first, beacons, paes);

	// optional -- fill the archive 
	for (i=1; i<paes->maxArchive; i++) {
		current = new_sol(beacons, paes);
		paes_update_grid(current, paes);
		paes_archive_soln(current, beacons, paes);
		free_sol(current);
    	} 
#ifdef DEBUG
	printf("PAES: gens %d archive %d depth %d\n", paes->gens, paes->maxArchive, paes->depth); 
	printf("starting with an archive of size %d\n", paes->inArchive);
	printf("initial front in pfront0.out\n");
	fp = fopen("pfront0.out", "w");
	int j;
	for(i=0; i<paes->inArchive; i++) {
		for(j=0; j<paes->objs; j++) fprintf(fp, "%2.3f ", paes->archive[i]->obj[j]);
		fprintf(fp, "\n");
	}
	fclose(fp);
#endif
	*ppaes = paes;
	return first;
}

// the main program - 
// utilisation : ./paesbeacons paes.dat beacons.dat
int main(int argc, char **argv)
{
	int i, j, result;
	solution_t *c, *m; // c is current solution, m is mutated solution
	int keepc; // keep c on next generation or replace it with m ?
	long t; // time measurement

	char *paramfile = argv[1]; // first argumeznt is the PAES parameter file
	beacons_t *beacons;
	paes_params_t *paes;

	if (argc != 3) goto usage;
	t = msectime();

	beacons = read_beacons(argv[2]); // second arg is the Beacon problem data beacons.dat
	c = paes_init(paramfile, &paes, beacons); // (c is current solution) first solution initialized
	for (i = 0; i < paes->gens; i++) {
		keepc = 1;
		m = mutate_sol(c, beacons, paes);

		result = compare_min(c->obj, m->obj, paes->objs);
		if (result != 1) { // c dominates m
			if (result == -1) { // m dominates c
				paes_update_grid(m, paes);
				paes_archive_soln(m, beacons, paes);
				keepc = 0;
			}
			else if(result == 0) { // non dominating solutions m and c
				result = paes_compare_to_archive(m, paes);
				if (result != -1) { // not dominated by archive
					paes_update_grid(m, paes);
					paes_archive_soln(m, beacons, paes);
					// if m in a less crowded area or m dominates archive
					if((paes->grid_pop[m->grid_loc] <= paes->grid_pop[c->grid_loc])||
                                                        (result==1)) {
						keepc = 0;
					}
				}
			}
		}

		// update current solution 
		if (keepc == 1)  
			free_sol(m);
		else {
			free_sol(c);
			c = m;
		}
	}

	t = msectime() - t;
	printf("runtime (secs) %1.3lf\n", t*1.0/1000L);

	// print the Pareto front
	FILE *fp;
	fp = fopen("pfront.out", "w");
	for(i=0; i<paes->inArchive; i++) {
		for(j=0; j<paes->objs; j++) fprintf(fp, "%2.3f ", paes->archive[i]->obj[j]);
		fprintf(fp, "\n");
	}
	fclose(fp);
	// print the Pareto set
	fp = fopen("pset.out", "w");
	for(i=0; i<paes->inArchive; i++)
		print_solFP(fp, paes->archive[i], beacons, paes);
	fclose(fp);

	printf("final archive of %d sols\n", paes->inArchive);
	printf("final front in pfront.out\n");
	printf("final set in pset.out\n");
	
	system("gnuplot -e ' set border 4095; set ticslevel 0;set xlabel \"tour length\"; set ylabel \"%% profit loss\"; set terminal pdf color; set output \"pfront.pdf\"; plot \"pfront.out\" u 1:2 with points lw 2 ps 2 t \"front\", \"pfront0.out\" u 1:2 with points lw 2 ps 2 t \"init\"'");
	printf("front plotted in pfront.pdf\n" );

// post optimisation, apply 2-opt
/*
	int nbImproved = 0;
	for(i=0; i<paes->inArchive; i++) {
		m = localsearch_sol(paes->archive[i], beacons, paes);
		if (m != NULL) {
			paes_update_grid(m, paes);
			paes_archive_soln(m, beacons, paes);
			free_sol(m);
			nbImproved++;
		}
			
	}

	// print the new Pareto front
	fp = fopen("pfront2.out", "w");
	for(i=0; i<paes->inArchive; i++) {
		for(j=0; j<paes->objs; j++) fprintf(fp, "%2.3f ", paes->archive[i]->obj[j]);
		fprintf(fp, "\n");
	}
	fclose(fp);
	// print the new Pareto set
	fp = fopen("pset2.out", "w");
	for(i=0; i<paes->inArchive; i++)
		print_solFP(fp, paes->archive[i], beacons, paes);
	fclose(fp);
	system("gnuplot -e ' set border 4095; set ticslevel 0;set xlabel \"tour length\"; set ylabel \"%% profit loss\"; set terminal pdf color; set output \"pfront2.pdf\"; plot \"pfront2.out\" u 1:2 with points lw 2 ps 2 t \"2opt front\", \"pfront.out\" u 1:2 with points lw 2 ps 2 t \"front\", \"pfront0.out\" u 1:2 with points lw 2 ps 2 t \"init\"'");
	printf("front plotted in pfront.pdf\n" );
*/

// drawing all solutions
//	draw_sol(0, beacons, paes, 0); // the map itself in beacons.pdf
/*
	for(i=0; i<paes->inArchive; i++) {
		char fname[128];
		sprintf(fname, "beacons.%d.pdf", i);
		draw_sol(paes->archive[i], beacons, paes, fname);
	}
	printf("solutions plotted in beacons.*.pdf\n" );
*/

	// free space
	free_sol(c);
	for (i=0; i<paes->inArchive; i++) free_sol(paes->archive[i]);
	free(paes->archive);
	free(paes->gl_offset);
	free(paes->gl_range);
	free(paes->gl_largest);
	free(paes->grid_pop);
	free(paes);

	for (i=0; i<beacons->nbBeacons; i++) free(beacons->distances[i]);
	free(beacons->distances);
	free(beacons->urgency);
	free(beacons->lon);
	free(beacons->lat);
	free(beacons);

	return 0;

usage:
	printf("%s <param file> <beacons file> -- syntax:\n", argv[0]);
	printf("param file syntax:\n");
	printf("seed <int seed value>\n");
	printf("gens <iterations>\n");
	printf("archive <archive size>\n");
	printf("depth <grid paes->depth -- eg 4>\n");

	return 0;
}

// read data -- see beacons.dat for format
beacons_t *read_beacons(char *beaconsfile)
{
	beacons_t *beacons;
	int i, j, distance;
	FILE *fp = fopen(beaconsfile, "r");
	if (!fp) { printf("no TSP file\n"); exit(0); }

	beacons = (beacons_t *)malloc(sizeof(beacons_t));
	if (!beacons) { printf("TSP malloc error\n"); exit(0); }
	fscanf(fp, "%d", &(beacons->nbBeacons));

	beacons->urgency = (int *)malloc(beacons->nbBeacons*sizeof(int));
	for(i=0; i<beacons->nbBeacons; i++)
		fscanf(fp, "%d", beacons->urgency + i);

	beacons->distances = (int **)malloc(beacons->nbBeacons*sizeof(int *));
	for(i=0; i<beacons->nbBeacons; i++)
		beacons->distances[i] = (int *)malloc(beacons->nbBeacons*sizeof(int));
	for(j=1; j<beacons->nbBeacons; j++) 
		for(i=0; i<j; i++) {
			fscanf(fp, "%d", &distance);
			beacons->distances[i][j] = beacons->distances[j][i] = distance;
		}
	// locations
	beacons->lat = (double *)malloc(beacons->nbBeacons*sizeof(double));
	beacons->lon = (double *)malloc(beacons->nbBeacons*sizeof(double));
	int index;
	for(i=0; i<beacons->nbBeacons; i++)
		fscanf(fp, "%d%lf%lf", &index, beacons->lat+i, beacons->lon+i);
	fclose(fp);

	return beacons;
}

// sols core functions
// all functions are valid, but they can be improved, eventually
solution_t *new_sol(beacons_t *beacons, paes_params_t *paes)
{
	int i;
	solution_t *sol = dup_sol((solution_t *)0, beacons, paes);

	int selected[beacons->nbBeacons];
	for(i=0; i<beacons->nbBeacons; i++)
		selected[i] = 0;

	// create a random tour (solution)
	int nbBeacons = rand() % beacons->nbBeacons; // nb beacons in this solution
	if (nbBeacons == 0) nbBeacons = 1;
	for(i=0; i<nbBeacons; i++) {
		int next = rand() % beacons->nbBeacons;
		while (selected[next]) { next++; if (next == beacons->nbBeacons) next = 0; }
		selected[next] = 1;
		sol->chrom[i] = next;
	}
	for(i=nbBeacons; i<beacons->nbBeacons; i++)  
		sol->chrom[i] = -1; // all not used tour slots invalidated

	evaluate_sol(sol, beacons);
	return sol;
}

// please compute the solution's values (sol->obj[_LENGTH], sol->obj[_URGENCYLOSS])
int evaluate_sol(solution_t * sol, beacons_t *beacons)
{
        int i, j, first;
        static int maxUrgency = 0;

	// in order to build a min function, 
	// max Sum_over_selected_beacons(urgency) 
	// <=>
	// min ( (maxUrgency=Sum_over_all_beacons(urgency)) - Sum_over_selected_beacons(urgency) ) / maxUrgency
        if (maxUrgency == 0) {
                for(i=0; i<beacons->nbBeacons; i++)
                        maxUrgency += beacons->urgency[i];
        }

        first = i = -1;
        sol->obj[_LENGTH] = 0.0;
        sol->obj[_URGENCYLOSS] = maxUrgency;
        for(j=0; j<beacons->nbBeacons; j++) {
                if (sol->chrom[j] == -1) continue;
                sol->obj[_URGENCYLOSS] -= beacons->urgency[sol->chrom[j]];
                if (first == -1) { i = first = j; continue; }
                sol->obj[_LENGTH] +=  beacons->distances[sol->chrom[i]][sol->chrom[j]];
                i = j;
        }

	sol->obj[_URGENCYLOSS] = 100.0 * sol->obj[_URGENCYLOSS] / maxUrgency;
        if (first != -1 && i != -1)
                sol->obj[_LENGTH] +=  beacons->distances[sol->chrom[first]][sol->chrom[i]];
        return 1; // evaluation done OK
}

// mutation: 4 possible transformations of beacons in a tour
solution_t *mutate_sol(solution_t * sol, beacons_t *beacons, paes_params_t *paes)
{
	solution_t *mute = dup_sol(sol, beacons, paes);
	
	// TP TSP MC : coder une mutation pour le TSP
	
	enum { _swap, _add, _delete, _replace} mode = rand() % 4; // swap or add or delete or replace nbBeacons

	int i;
	int nbBeacons = 0;
	int selected[beacons->nbBeacons];
	for(i=0; i<beacons->nbBeacons; i++) selected[i] = 0;
	for(i=0; i<beacons->nbBeacons; i++) 
		if (sol->chrom[i] != -1) { 
			selected[sol->chrom[i]] = 1; 
			nbBeacons++; 
	}

	if (nbBeacons == 0) mode = _add; // add to empty tour
	else if (nbBeacons == 1 && (mode == _swap || mode == _delete)) mode = _replace; // replace instead of swap or delete  
	else if (nbBeacons == beacons->nbBeacons && (mode == _add || mode == _replace)) 
		mode = rand()%2 ? _swap : _delete; // can't add or replace, swap or delete instead
	
	if (mode == _swap) { 
		int site1 = rand() % beacons->nbBeacons;
		while (sol->chrom[site1] == -1) { site1++; if (site1 == beacons->nbBeacons) site1 = 0; }
		int site2 = rand() % beacons->nbBeacons;
		while (sol->chrom[site2] == -1 || site2 == site1) { site2++; if (site2 == beacons->nbBeacons) site2 = 0; }
		int tmp = sol->chrom[site1];
		mute->chrom[site1] = sol->chrom[site2];
		mute->chrom[site2] = sol->chrom[site1];
	} else if (mode == _add) { 
		int site = rand() % beacons->nbBeacons;
		while (sol->chrom[site] != -1) { site++; if (site == beacons->nbBeacons) site = 0; }
		int newtown = rand() % beacons->nbBeacons;
		while (selected[newtown]) { newtown++; if (newtown == beacons->nbBeacons) newtown = 0; }
		mute->chrom[site] = newtown;
	} else if (mode == _delete) { 
		int site = rand() % beacons->nbBeacons;
		while (sol->chrom[site] == -1) { site++; if (site == beacons->nbBeacons) site = 0; }
		mute->chrom[site] = -1;
	} else if (mode == _replace) { 
		int site = rand() % beacons->nbBeacons;
		while (sol->chrom[site] == -1) { site++; if (site == beacons->nbBeacons) site = 0; }
		int newtown = rand() % beacons->nbBeacons;
		while (selected[newtown]) { newtown++; if (newtown == beacons->nbBeacons) newtown = 0; }
		mute->chrom[site] = newtown;

	}
		
	evaluate_sol(mute, beacons);

	return mute;
}

// print objectives (length %loss) and then solution  (beacons selected)
// into a file, use stdout for printing onto terminal
void print_solFP(FILE *fp, solution_t *sol, beacons_t *beacons, paes_params_t *paes)
{
	int i, j;
	assert(sol != NULL);
	// for(i=0; i<paes->objs; i++) fprintf(fp, "%d ", (int)sol->obj[i]);
	fprintf(fp, "%5d %2.2lf ", (int)sol->obj[_LENGTH], sol->obj[_URGENCYLOSS]);
	for(i=0; i<beacons->nbBeacons; i++) 
		if (sol->chrom[i] != -1) 
			fprintf(fp, " %d ", sol->chrom[i]);
	fprintf(fp, "\n");
}

// alloc or dup a solution
solution_t *dup_sol(solution_t *sol, beacons_t *beacons, paes_params_t *paes)
{
	solution_t *dup = (solution_t *)malloc(sizeof(solution_t));

	dup->chrom = (int *)malloc(beacons->nbBeacons*sizeof(int));
	dup->obj = (double *)malloc(paes->objs*sizeof(double));
	if (sol) {
		memcpy(dup->chrom, sol->chrom, beacons->nbBeacons*sizeof(int));
		memcpy(dup->obj, sol->obj, paes->objs*sizeof(double));
		dup->grid_loc = sol->grid_loc;
	}
	return dup;
}

void free_sol(solution_t *sol)
{
	free(sol->chrom);
	free(sol->obj);
	free(sol);
}
// ========================================================================================
// PAES core functions

int paes_compare_to_archive(solution_t *s, paes_params_t *paes)  // compares a PAES_solution to every member 
                                                            // of the PAES_ArcMax. Returns -1 if dominated by
{                                                           // any member, 1 if dominates any member, and 0 otherwise
	int i=0;
	int result=0;

	while((i<paes->inArchive)&&(result!=1)&&(result!=-1)) {
		result = compare_min(s->obj, paes->archive[i]->obj, paes->objs);
		i++;
	}

	return(result);
}
      

void add_to_archive(solution_t *s, beacons_t *beacons, paes_params_t *paes)
{
	paes->archive[paes->inArchive] = dup_sol(s, beacons, paes);
	paes->inArchive++;
}

int compare_min(double *first, double *second, int n)
{
	// compares two n-dimensional vectors of objective values for minimization problems
	// returns 1 if first dominates second,
	// -1 if second dominates first, and 0 otherwise

	int obj = 0;
	int deflt = 0;
	int current;

	do 
	{
		if(*first < *second)
			current = 1;
		else if(*second < *first)
			current = -1;
		else
			current = 0;
		if((current)&&(current==-deflt)) {

			return(0);
		}
		if(current!=0) {
			deflt = current;
		}
		obj++;
		*first++;
		*second++;
	} while(obj < n);

	return(deflt);
}

int paes_equal(double *first, double *second, int n)
{
	// checks to n-dimensional vectors of paes->objs to see if they are identical
	// returns 1 if they are, 0 otherwise

	int obj = 0;

	do
	{
		if(*first!=*second)
			return(0);
		*first++;
		*second++;
		obj++;
		// printf("%d\n",obj);
	}
	while(obj < n);
	return(1);
}


void paes_archive_soln(solution_t *s, beacons_t *beacons, paes_params_t *paes)
{
	// given a PAES_solution s, add it to the PAES_ArcMax if
	// a) the PAES_ArcMax is empty 
	// b) the PAES_ArcMax is not full and s is not dominated or paes_equal to anything currently in the PAES_ArcMax
	//  s dominates anything in the PAES_ArcMax                        
	// d) the PAES_ArcMax is full but s is nondominated and is in a no more crowded square than at least one PAES_solution
	// in addition, maintain the PAES_ArcMax such that all solutions are nondominated. 

	int i;
	int repl;
	int yes = 0;
	int most;
	int result;
	int join = 0;
	int old_arclength;
	int set = 0;


	int tag[paes->maxArchive];


	for (i = 0; i < paes->maxArchive; i++) { tag[i] = 0; }


	if (paes->inArchive == 0) { add_to_archive(s, beacons, paes); return; }


	i = 0;
	result = 0;
	while((i < paes->inArchive)&&(result!=-1)) {
		result = paes_equal(s->obj, paes->archive[i]->obj, paes->objs);
		if (result == 1) break;
		result = compare_min(s->obj, paes->archive[i]->obj, paes->objs);

		if ((result == 1) && (join == 0)) {
			free_sol(paes->archive[i]);
			paes->archive[i] = dup_sol(s, beacons, paes);
			join = 1;
		} else if (result == 1) {
			tag[i] = 1;
			set = 1;
		}	    	    
		i++;
	}

	old_arclength = paes->inArchive;
	if (set==1) {
		solution_t **tmp;

		if (!(tmp = (solution_t **)malloc(paes->inArchive*sizeof(solution_t **)))) {
			perror("Out of memory -- aborting");
			exit(-1);
		}
		for (i = 0; i < paes->inArchive; i++) { tmp[i] = paes->archive[i]; }
		paes->inArchive = 0;

		for (i = 0; i < old_arclength; i++) {
			if (tag[i]!=1) {
				paes->archive[paes->inArchive]=tmp[i];
				paes->inArchive++;
			}
			else {
				free_sol(paes->archive[i]);
			}
		}
		free(tmp);
	}


	if ((join==0)&&(result==0)) { // ie PAES_solution is non-dominated by the list
		if (paes->inArchive == paes->maxArchive) {	  
			most = paes->grid_pop[s->grid_loc];
			for (i = 0; i < paes->inArchive; i++) {
				if (paes->grid_pop[paes->archive[i]->grid_loc] > most) {
					most = paes->grid_pop[paes->archive[i]->grid_loc];
					repl = i;
					yes = 1;
				}
			}
			if (yes) {
				free_sol(paes->archive[repl]);
				paes->archive[repl] = dup_sol(s, beacons, paes);	     
			}
		} else {
			add_to_archive(s, beacons, paes);
		}
	}
}

int paes_find_loc(double *eval, paes_params_t *paes)
{
	// find the grid location of a PAES_solution given a vector of its objective values

	int loc = 0;
	int d;
	int n = 1;

	int i;

	int inc[paes->objs];
	double width[paes->objs];

	// if the PAES_solution is out of range on any objective, return 1 more than the maximum possible grid location number
	for (i = 0; i < paes->objs; i++) {
		if ((eval[i] < paes->gl_offset[i])||(eval[i] > paes->gl_offset[i] + paes->gl_range[i]))
			return((int)pow(2,(paes->objs*paes->depth)));
	}

	for (i = 0; i < paes->objs; i++) {
		inc[i] = n;
		n *=2;
		width[i] = paes->gl_range[i];
	}

	for (d = 1; d <= paes->depth; d++) {
		for (i = 0; i < paes->objs; i++) {
			if(eval[i] < width[i]/2+paes->gl_offset[i])
				loc += inc[i];
			else
				paes->gl_offset[i] += width[i]/2;
		}
		for (i = 0; i < paes->objs; i++) {
			inc[i] *= (paes->objs *2);
			width[i] /= 2;
		}
	}
	return(loc);
} 

void paes_update_grid(solution_t *s, paes_params_t *paes)
{
	// recalculate ranges for grid in the light of a new PAES_solution s
	static int change = 0;
	int a, b;
	int square;
	double offset[paes->objs];
	double largest[paes->objs];
	double sse;
	double product;
	int bound = pow(2, (paes->objs*paes->depth));

	for (a = 0; a < paes->objs; a++) {
		offset[a] = PAES_LARGE;
		largest[a] = -PAES_LARGE;
	}

	for (b = 0; b < paes->objs; b++) {
		for (a = 0; a < paes->inArchive; a++)	{
			if (paes->archive[a]->obj[b] < offset[b])
				offset[b] = paes->archive[a]->obj[b];
			if (paes->archive[a]->obj[b] > largest[b])
				largest[b] = paes->archive[a]->obj[b];		   
		}
	}

	for (b = 0; b < paes->objs; b++) {
		if (s->obj[b] < offset[b])
			offset[b] = s->obj[b];
		if (s->obj[b] > largest[b])
			largest[b] = s->obj[b];		   
	}

	sse = 0.0;
	product = 1.0;

	for (a = 0; a < paes->objs; a++) {
		sse += ((paes->gl_offset[a] - offset[a])*(paes->gl_offset[a] - offset[a]));
		sse += ((paes->gl_largest[a] - largest[a])*(paes->gl_largest[a] - largest[a]));
		product *= paes->gl_range[a]; 
	}

	// printf("sse = %f\n", sse);
	if (sse > (0.1 * product * product))	//if the summed squared error (difference) between old and new
		//minima and maxima in each of the paes->objs
	{                                   //is bigger than 10 percent of the square of the size of the space
		change++;                         // then renormalise the space and recalculte grid locations

		for (a = 0; a < paes->objs; a++) {
			paes->gl_largest[a] = largest[a]+0.2*largest[a];
			paes->gl_offset[a] = offset[a]+0.2*offset[a];
			paes->gl_range[a] = paes->gl_largest[a] - paes->gl_offset[a];
		}

		//for (a = 0; a < pow(2, (paes->objs*paes->depth)); a++) {
		for (a = 0; a < bound; a++) {
			paes->grid_pop[a] = 0;
		}

		for (a = 0; a < paes->inArchive; a++) {
			square = paes_find_loc(paes->archive[a]->obj, paes);
			paes->archive[a]->grid_loc = square;
			paes->grid_pop[square]++;
		}
	}
	square = paes_find_loc(s->obj, paes);
	s->grid_loc = square;
	paes->grid_pop[bound] = -5;
	paes->grid_pop[square]++;   

}


// time
#include <sys/time.h>
long msectime()
{
	struct timeval tv;

	gettimeofday(&tv, (struct timezone *)0);

	return (1000L*tv.tv_sec) + (((long) tv.tv_usec)/1000L);
}

// drawing beacons map 
void draw_sol(solution_t *sol, beacons_t *beacons, paes_params_t *paes, char *fname)
{
	int i;
	FILE *fp = fopen("beaconsMap.dat", "w");
	for(i=0; i<beacons->nbBeacons; i++)
		fprintf(fp, "%d %2.3lf %2.3lf %d\n", i, beacons->lon[i], beacons->lat[i], beacons->urgency[i]);
	fclose(fp);

	if (sol == NULL) {
		system("gnuplot -e ' set key noautotitle; set border 4095; set terminal pdf color; set output \"beaconsMap.pdf\"; plot \"beaconsMap.dat\" u 2:3:4 with points palette lw 2 ps 2, \"beaconsMap.dat\" u ($2):($3)-0.7:1 with labels lw 2'");
		if (fname != NULL) rename("beaconsMap.pdf", fname);
//		printf("Beacons map  plotted in %s\n", fname ? fname : "beacons.pdf");
	}
	else {
		fp = fopen("solTrace.dat", "w");
		int firstb = -1;
		for(i=0; i<beacons->nbBeacons; i++) {
			int b = sol->chrom[i];
			if(b != -1) {
				if (firstb == -1) firstb = b;
				fprintf(fp, "%2.3lf %2.3lf\n", beacons->lon[b], beacons->lat[b]);
			}
		}
		fprintf(fp, "%2.3lf %2.3lf\n", beacons->lon[firstb], beacons->lat[firstb]);
		fclose(fp);
		char cmd[1024];
		sprintf(cmd, "gnuplot -e ' set border 4095; set terminal pdf color; set output \"beacons.pdf\"; plot \"beaconsMap.dat\" u 2:3:4 with points palette lw 2 ps 2 t \"beacons\", \"solTrace.dat\" u 1:2 with lines lw 2 t \"sol %d %2.2lf%%\"'", (int)sol->obj[_LENGTH], sol->obj[_URGENCYLOSS]);
		system(cmd);
		//system("gnuplot -e ' set border 4095; set terminal pdf color; set output \"beacons.pdf\"; plot \"beaconsMap.dat\" u 2:3:4 with points palette lw 2 ps 2 t \"beacons\", \"solTrace.dat\" u 1:2 with lines lw 2 t \"sol\"'");
		if (fname != NULL) rename("beacons.pdf", fname);
//		printf("solution below plotted in beacons.pdf\n" );
//		printf("solution plotted in %s\n", fname ? fname : "beacons.pdf");
//		print_sol(sol, beacons, paes);
		unlink("solTrace.dat");
	}
	unlink("beaconsMap.dat");
}

// do a local search on the solution, optimizing the tour length
// return the modified solution if improved, else NULL
solution_t *localsearch_sol(solution_t *sol, beacons_t *beacons, paes_params_t *paes)
{
	int i, j;
	int nbBeacons = 0;
	solution_t *mute = dup_sol(sol, beacons, paes);
	for(i=0; i<beacons->nbBeacons; i++) 
		if (mute->chrom[i] != -1) 
			mute->chrom[nbBeacons++] = mute->chrom[i]; 

	int better = 0;
	for (i=0; i<nbBeacons; i++) 
		for (j=0; j<nbBeacons; j++) {
			if ((j==(i-1)) || (j==i) || (j==(i+1)) || 
				((j==(nbBeacons-1)) && (i==0)) || 
				((i==(nbBeacons-1)) && (j==0)))
				continue;
			int t1 = mute->chrom[i];
			int t2 = mute->chrom[(i+1)%nbBeacons];
			int t3 = mute->chrom[j];
			int t4 = mute->chrom[(j+1)%nbBeacons];
			
			if ((beacons->distances[t1][t2] + beacons->distances[t3][t4]) > 
				(beacons->distances[t1][t3] + beacons->distances[t2][t4])) {
				mute->chrom[(i+1)%nbBeacons] = t3;
				mute->chrom[j] = t2;
				better = 1;
			}
		}
	if (better) {
		evaluate_sol(mute, beacons);
		return mute;
	}
	else {
		free_sol(mute);
		return (solution_t *)0;
	}
}

